
from unittest import TestCase

from namedspace import namedspace

class NamedspaceTests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.mock_default_value = "Mock default value"
        cls.mock_name_template = "Mock name for {id}"
        cls.mock_default_factory = staticmethod(lambda ns: cls.mock_name_template.format(id=ns.id))
        cls.mock_id = "mock_id"

        cls.TestNamedspace1 = namedspace("TestNamedspace1", "id", optional_fields=("name", "description", "extra"),
                mutable_fields=("name", "description"), default_values={"extra": cls.mock_default_value},
                default_value_factories={"name": cls.mock_default_factory})
        cls.test_ns1 = cls.TestNamedspace1(id=cls.mock_id)

        cls.TestNamedspace2 = namedspace("TestNamedspace2", "id", optional_fields="name",
                default_value_factories={"name": cls.mock_default_factory})
        cls.test_ns2 = cls.TestNamedspace2(id=cls.mock_id)

    def test_missing_required_fields(self):
        """
        A ValueError should be raised when a namedspace class is
        instantiated without values for required fields.
        """
        self.assertRaises(ValueError, self.TestNamedspace1)

    def test_required_fields(self):
        """
        Values passed as values for required fields when a
        namedspace class is instantiated should be available as
        attributes on the instance.
        """
        self.assertIs(self.test_ns1.id, self.mock_id)

    def test_mutable_fields(self):
        """
        Fields designated as mutable should be modifiable.
        """
        self.assertRaises(AttributeError, lambda: self.test_ns1.description)

        new_value = "new_value"
        self.test_ns1.description = new_value
        self.assertIs(self.test_ns1.description, new_value)

        del self.test_ns1.description

        self.assertRaises(AttributeError, lambda: self.test_ns1.description)

    def test_immutable_fields(self):
        """
        Fields not designated as mutable should be read-only.
        """
        self.assertRaises(self.TestNamedspace1.ReadOnlyFieldError,
                lambda: setattr(self.test_ns1, "extra", "new value"))

    def test_immutable_namedspace(self):
        """
        Namedspaces that have no mutable fields should consider themselves read-only.
        """
        self.assertRaises(self.TestNamedspace2.ReadOnlyNamedspaceError,
                lambda: setattr(self.test_ns2, "name", "new name"))

    def test_namedspace_hashable(self):
        """
        Immutable namedspaces should be hashable. Mutable namedspaces should not be.
        """
        self.assertRaises(self.TestNamedspace1.MutableNamedspaceError, lambda: hash(self.test_ns1))

        ns2_hash = hash(self.test_ns2)
        self.assertIsInstance(ns2_hash, int)

    def test_default_values(self):
        """
        Default values for fields should be retrieved (if they exist)
        from the default_values mapping provided when the namedspace
        class was generated.
        """
        self.assertIs(self.test_ns1.extra, self.mock_default_value)

    def test_default_value_factories(self):
        """
        Default values for fields should be retrieved by acquiring
        the default factory for a field (if it exists) from the
        default_value_factories mapping provided when the namedspace
        class was generated, calling the factory and passing the
        namedspace instance to it, and returning the resulting value.
        """
        self.assertEqual(self.test_ns1.name, self.mock_name_template.format(id=self.test_ns1.id))

    def test_field_names(self):
        """
        The _field_names property should return the correct values.
        """
        self.assertEqual(self.test_ns1._field_names, frozenset(("id", "name", "description", "extra")))

    def test_field_values(self):
        """
        The _field_values property should return the correct values.
        """
        self.assertEqual(frozenset(self.test_ns1._field_values), frozenset((
                self.test_ns1.id,
                self.test_ns1.name,
                getattr(self.test_ns1, "description", None),
                self.test_ns1.extra,
            )))

    def test_field_items(self):
        """
        The _field_items property should return the correct values.
        """
        self.assertEqual(self.test_ns1._field_items, frozenset((
                ("id", self.test_ns1.id),
                ("name", self.test_ns1.name),
                ("description", getattr(self.test_ns1, "description", None)),
                ("extra", self.test_ns1.extra),
            )))


class SubNamedspace(namedspace("_SubNamedspace", ("id", "name"))):
    _name_tmpl = "Overridden name for {id}"

    @property
    def name(self):
        return self._name_tmpl.format(id=self.id)


class SubNamedspaceTests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.mock_id = "mock_id"
        cls.mock_name = "mock_name"
        cls.test_ns1 = SubNamedspace(id=cls.mock_id, name=cls.mock_name)

    def test_overridden_property(self):
        """
        Properties with the same name as namedspace fieldnames should
        override the field values.
        """
        self.assertEqual(self.test_ns1.name, SubNamedspace._name_tmpl.format(id=self.mock_id))

    def test_overridden_values(self):
        """
        Overridden property values should properly appear in value collections.
        """
        self.assertIn(SubNamedspace._name_tmpl.format(id=self.mock_id), self.test_ns1._field_values)




import sys as _sys

from collections import Container
from collections import Mapping
from collections import Hashable
from collections import MutableMapping

from frozendict import frozendict


def namedspace(typename, required_fields=(), optional_fields=(), mutable_fields=(), default_values=frozendict(),
               default_value_factories=frozendict()):
    """Builds a new namedspace class.

    >>> from itertools import count

    >>> Someclass = namedspace("Someclass", "id", optional_fields=("name", "description", "extra"),
    ...     mutable_fields=("description", "extra"), default_values={"description": "None available"},
    ...     default_value_factories={"id": lambda self, counter=count(start=1): counter.next(),
    ...         "name": lambda self: "Name for id={id}".format(id=self.id)})

    >>> sc1 = Someclass()

    >>> sc1.id
    1

    The value of 1 was automatically assigned by the
    default_value_factory for the 'id' field.

    >>> sc1.name
    'Name for id=1'

    This value was also generated by a default value factory, based on
    the value of another field.

    >>> sc1.description
    'None available'

    This value came from the default_values mapping.

    >>> sc1.description = "Some fancy description"
    >>> sc1.description
    'Some fancy description'

    The description field was set as a mutable field, which allows
    it to be modified.

    >>> del sc1.description

    Its value can also be deleted.

    >>> sc1.description
    'None available'

    Since its modified value was deleted, and it has a default value,
    it has reverted to its default value.

    >>> sc1.extra
    Traceback (most recent call last):
        <snip/>
    AttributeError: "Field 'extra' does not yet exist in this Someclass namedspace instance."

    The extra field is a valid field in this namedspace, but it has not
    yet been assigned a value and does not have a default. Since it one
    of the mutable fields, we can give it a value.

    >>> sc1.extra = "Extra cool value"
    >>> sc1.extra
    'Extra cool value'

    Only fields that have been declared as either required or optional
    are allowed.

    >>> sc1.some_other_field = "some other value"
    Traceback (most recent call last):
        <snip/>
    FieldNameError: "Field 'some_other_field' does not exist in Someclass namedspace."

    If we instantiate another instance, our id field will get the next counter value.
    >>> sc2 = Someclass()
    >>> sc2.id
    2

    Of course we can always supply all of the values at instantiation time.

    >>> sc3 = Someclass(id=1234, name="Foo", description="Bar", extra="Credit")

    All of the built in methods are "protected" so as to not pollute
    the public namespace of each instance.

    >>> sc3._field_names
    frozenset(['description', 'id', 'name', 'extra'])

    >>> sc3._field_values
    ('Bar', 1234, 'Foo', 'Credit')

    >>> sc3._field_items
    set([('description', 'Bar'), ('extra', 'Credit'), ('id', 1234), ('name', 'Foo')])

    """

    # Initialize the list of arguments that will get put into the
    # doc string of the generated class
    arg_list_items = []

    #
    # Validate parameters
    #
    for arg_name in ("required_fields", "optional_fields", "mutable_fields"):
        arg_value = locals()[arg_name]

        if isinstance(arg_value, basestring):
            arg_value = (arg_value,)
            exec "{arg_name} = arg_value".format(arg_name=arg_name)
        elif not isinstance(arg_value, Container):
            raise ValueError("Value for argument '{arg_name}' must be a string or container of strings.".format(
                    arg_name=arg_name))

        for field_name in arg_value:
            if not isinstance(field_name, basestring):
                raise ValueError("Items of container argument '{arg_name}' must be strings.".format(arg_name=arg_name))

        if len(arg_value) != len(frozenset(arg_value)):
            raise ValueError("Value for argument '{arg_name}' contains duplicate fields.".format(
                    arg_name=arg_name))

        arg_list_items.append("{arg_name}={arg_value!r}".format(arg_name=arg_name, arg_value=tuple(arg_value)))

        exec "{arg_name} = frozenset(arg_value)".format(arg_name=arg_name)

    all_fields = required_fields.union(optional_fields)

    for field_name in mutable_fields:
        if field_name not in all_fields:
            raise ValueError("Mutable field '{field_name}' is not a required or optional field.".format(
                    field_name=field_name))

    for arg_name in ("default_values", "default_value_factories"):
        arg_value = locals()[arg_name]
        if not isinstance(arg_value, Mapping):
            raise ValueError("Value for argument '{arg_name}' must be a mapping.".format(arg_name=arg_name))

        default_field_names = frozenset(arg_value.iterkeys())
        if not default_field_names.issubset(all_fields):
            bad_default_field_names = default_field_names - all_fields
            raise ValueError("Value for argument '{arg_name}' contains invalid field(s) '{field_names}'.".format(
                    arg_name=arg_name, field_names=", ".join(bad_default_field_names)))

        arg_list_items.append("{arg_name}={arg_value!r}".format(arg_name=arg_name, arg_value=dict(arg_value)))

        exec "{arg_name} = frozendict(arg_value)".format(arg_name=arg_name)

    for field_name, factory in default_value_factories.iteritems():
        if not callable(factory):
            raise ValueError("Default value factory for '{field_name}' is not callable.".format(field_name=field_name))

    # Fill-in the class template
    class_definition = _class_template.format(
        typename=typename,
        arg_list=", ".join(arg_list_items),
        )

    # Execute the template string in a temporary namespace and support
    # tracing utilities by setting a value for frame.f_globals['__name__']
    namespace = dict(
        __name__='namedspace_{typename}'.format(typename=typename),
        all_fields=all_fields,
        required_fields=required_fields,
        mutable_fields=mutable_fields,
        default_values=default_values,
        default_value_factories=default_value_factories,
        Hashable=Hashable,
        MutableMapping=MutableMapping,
        )

    #
    # Code from here down copied verbatim from namedtuple
    #
    try:
        exec class_definition in namespace
    except SyntaxError as e:
        raise SyntaxError(e.message + ':\n' + class_definition)
    result = namespace[typename]

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    try:
        result.__module__ = _sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return result

_class_template = """\
class {typename}(object):
    "{typename}({arg_list})"

    class FieldNameError(AttributeError, KeyError): pass
    class ReadOnlyNamedspaceError(TypeError): pass
    class ReadOnlyFieldError(TypeError): pass
    class MutableNamedspaceError(TypeError): pass

    _all_fields = all_fields
    _required_fields = required_fields
    _mutable_fields = mutable_fields
    _default_values = default_values
    _default_value_factories = default_value_factories

    def __init__(self, **kwargs):
        self._field_value_storage = dict()

        for field_name, field_value in kwargs.iteritems():
            if field_name in self._all_fields:
                self._field_value_storage[field_name] = field_value
            else:
                raise ValueError("field '{{field_name}} does not exist in the {typename} namedspace.".format(
                        field_name=field_name))

        for field_name in self._all_fields:
            try:
                field_value = self._get_value(field_name)
            except self.FieldNameError:
                field_value = None

            if field_value in (None, "") and field_name in self._required_fields:
                raise ValueError("A value for field '{{field_name}}' is required.".format(field_name=field_name))
            elif not field_name in self._field_value_storage:
                self._field_value_storage[field_name] = field_value

    def __repr__(self):
        'Return a nicely formatted representation string'
        return '{typename}({{items!r}})'.format(items=self.as_dict())

    #
    # Generic value access methods
    #
    def _get_value(self, field_name):
        if not field_name in self._all_fields:
            raise self.FieldNameError("Field '{{field_name}}' does not exist in the {typename} namedspace.".format(
                    field_name=field_name))

        field_value = self._field_value_storage.get(field_name)
        if field_value is None:
            field_value = self._default_values.get(field_name)
            if field_value is None:
                factory = self._default_value_factories.get(field_name)
                if factory is None:
                    raise self.FieldNameError("Field '{{field_name}}' does not yet exist in this {typename}"
                            " namedspace instance.".format(field_name=field_name))
                else:
                    field_value = factory(self)

        return field_value

    def _validate_field_mutability(self, field_name):
        if not field_name in self._all_fields:
            raise self.FieldNameError("Field '{{field_name}}' does not exist in {typename} namedspace.".format(
                    field_name=field_name))
        if not field_name in self._mutable_fields:
            if self._mutable_fields:
                raise self.ReadOnlyFieldError("Field '{{field_name}}' of {typename} namedspace is read-only.".format(
                        field_name=field_name))
            else:
                raise self.ReadOnlyNamedspaceError("{typename} namedspace is read-only.")

    def _set_value(self, field_name, field_value):
        self._validate_field_mutability(field_name)
        self._field_value_storage[field_name] = field_value

    def _del_value(self, field_name):
        self._validate_field_mutability(field_name)
        del self._field_value_storage[field_name]

    #
    # Namedspace API
    #
    @property
    def _field_names(self):
        return self._all_fields

    @property
    def _field_names_iter(self):
        return iter(self._all_fields)

    @property
    def _field_values_iter(self):
        for field_name in self._all_fields:
            yield getattr(self, field_name, None)

    @property
    def _field_values(self):
        return tuple(self._field_values_iter)

    @property
    def _field_items_iter(self):
        for field_name in self._all_fields:
            yield (field_name, getattr(self, field_name, None))

    @property
    def _field_items(self):
        set_type = set if self._mutable_fields else frozenset
        return set_type(self._field_items_iter)

    @property
    def _as_dict(self):
        'Return a the namedspace values as a new dictionary.'
        return dict(self._field_items_iter)

    #
    # Attribute API
    #
    def __getattr__(self, attr_name):
        try:
            return self._get_value(attr_name)
        except self.FieldNameError as e:
            raise AttributeError(str(e))

    def __setattr__(self, attr_name, attr_value):
        if attr_name == "_field_value_storage":
            return super({typename}, self).__setattr__(attr_name, attr_value)
        else:
            return self._set_value(attr_name, attr_value)

    def __delattr__(self, attr_name):
        return self._del_value(attr_name)

    #
    # Hashable API
    #
    def __hash__(self):
        if self._mutable_fields:
            raise self.MutableNamedspaceError("Mutable {typename} namedspace instance is not hashable.")
        else:
            return hash(self._field_values)

    #
    # MutableMapping API
    #
    def __contains__(self, name):
        return name in self._field_names

    def __iter__(self):
        return self._field_names_iter

    def __len__(self):
        return len(self._field_values_iter)

    def __getitem__(self, item_name):
        try:
            return self._get_value(item_name)
        except self.FieldNameError as e:
            raise KeyError(str(e))

    def __setitem__(self, item_name, item_value):
        return self._set_value(item_name, item_value)

    def __delitem__(self, item_name):
        return self._del_value(item_name)


Hashable.register({typename})
MutableMapping.register({typename})
"""




import os

from setuptools import setup, find_packages

__version__ = "1.0.0"

def file_read(filename):
    filepath = os.path.join(os.path.dirname(__file__), filename)
    with open(filepath) as flo:
        return flo.read()

setup(
    name = "namedspace",
    version = __version__,
    packages = find_packages(),
    install_requires = ["frozendict"],
    author = "Warren A. Smith",
    author_email = "warren@wandrsmith.net",
    description = "Namespace class factory.",
    long_description = file_read("README.md"),
    license = "PSF",
    keywords = "namedspace namespace",
    url = "https://github.com/wsmith323/namedspace",
    test_suite = "namedspace.tests",
)
